; frmk_do.s
; by Mark Slagell

; emulates form_do() with the following differences:
;  >  any selectable object may be selected with a key
;  >  any keypress may be returned even if not associated with an object
;  >  after clicking one object you may change your mind and select another
;      without releasing the mouse button
;  >  there is no support for object editing

; call as frmk_do( OBJECT *tree, int *keysub );
; where keysub points to an integer array like this:
; { BUTTON1,'1','A','a',-1,
;   BUTTON2,'2','B','b',-1,
;   CANCEL,27,-1
;   -1 }

; The -1 value delimits each list of key subs, and an extra -1 signals
; the end of the entire list.  To assign a key without an ASCII value,
; give its scancode plus 128; for example, 196 for F10 or 226 for Help.

; The list may be terminated with -2 instead of -1. In this case any
; keypress not assigned to a button will be returned with bit 14 set;
; for example, the spacebar would return 16416 (32+16384) if it were not
; assigned to a selectable object somewhere in the list.  Note that this
; makes every unassigned key its own invisible exit object.

;register usage:
; d3: temporary data
; d4: object number
; d5: object flags
; d6: object state
; d7: object's next sibling
; a4: temporary address
; a5: tree address


.globl	_frmk_do

.globl	_evnt_multi
.globl	_objc_find
.globl	_objc_change
.globl	_graf_mkstate

	.bss
kp0:	.ds.l	1
shift:	.ds.w	1
clicks:	.ds.w	1
key:	.ds.w	1
but:	.ds.w	1
mx:	.ds.w	1
my:	.ds.w	1
	.text

_frmk_do:
	link	a6,#0
	move.l	8(a6),a5	; tree address
	move.l	12(a6),kp0
	
	movem.l	d3-d7/a4-a5,-(sp)

TOP:
	pea	clicks
	pea	key
	pea	shift
	pea	but
	pea	my
	pea	mx
	clr.l	-(sp)
	clr.l	-(sp)
	clr.l	-(sp)
	clr.l	-(sp)
	clr.l	-(sp)
	clr.l	-(sp)
	clr.l	-(sp)
	move.l	#$10001,-(sp)
	move.l	#$30002,-(sp)
	jsr	_evnt_multi
	add.w	#60,sp
	btst	#1,d0
	beq	KEY_PRESSED
BUTTON_PRESSED:
	bsr	find_ob
	tst	d4
	bmi	TOP		; we're outside dialog
	btst	#6,d5
	beq	sk0
	cmp	#1,clicks	; it's a touchexit!
	bls	ALLDONE
	or	#$8000,d4	; set high bit if doubleclicked
	bra	ALLDONE
sk0:
	btst	#0,d5		; go back if not selectable
	beq	TOP
	btst	#3,d6		; go back if disabled
	bne	TOP
	btst	#4,d5
	bne	RADIO		; radio logic
	
	bsr	inv_ob		; invert our object
	move	d4,d3		; remember which obj
lp0:	bsr	gm		; check mouse pos & button state
	btst	#0,but+1	; is left button down?
	beq	RELEASED	; no; declare the object chosen
	bsr	find_ob
	cmp	d3,d4		; still over the same obj?
	beq	lp0		; yes, keep watching
	move	d3,d4		; no, user abandoned the object
	bsr	inv_ob		; make it like it was
	bra	TOP
RELEASED:
	btst	#0,d6		; if button isn't selected,
	beq	TOP		; go to the beginning
;	bra	TEST_EXIT

TEST_EXIT:
	btst	#2,d5		; is it an exit object?
	beq	TOP		;  no
;	bra	ALLDONE		;  yes
ALLDONE:			; return d4
	move	d4,d0
FRCEXIT:
	movem.l	(sp)+,d3-d7/a4-a5
	unlk	a6
	rts

RADIO:	
	move.l	a5,a4
olp1:	move	2(a4),d0	; get index of first child
	bmi	lpend1		;  branch if there aren't any
ilp1:	cmp	d0,d4		; if child is the obj selected
	beq	rfound		; then a4 points to its parent
	cmp	4(a4),d0
	beq	lpend1		; exit if we've seen all sibs
	mulu	#24,d0
	move	0(a5,d0.l),d0	; else get index of next one
	bra	ilp1
lpend1:
	btst	#5,7(a4)	; check for lastob
	bne	no_radio_desel	; shouldn't get here (root=radio!?)
	add.w	#24,a4		; try next prospective parent
	bra	olp1
rfound:
	move	d4,d3
	move	2(a4),d4	; get index of first child
lp2:
	bsr	getstate
	cmp	d3,d4		; if it's not the one being selected,
	beq	lpend2
	btst	#4,d5		; and RBUTTON is indicated,
	beq	lpend2
	bsr	clr_ob		; deselect it
lpend2:
	cmp	4(a4),d4	; check for last sibling
	beq	out2
	move	d7,d4		; get next sibling
	bra	lp2
out2:
	move	d3,d4
no_radio_desel:
	bsr	set_ob		; show which obj was selected
	move	d4,d3
lp3:	bsr	gm		; check mouse pos & button state
	btst	#0,but+1
	beq	TEST_EXIT	; done if left button released
	bsr	find_ob
	cmp	d3,d4
	beq	lp3
	bra	TEST_EXIT	; also done if pointer moved off obj

KEY_PRESSED:
	move	key,d0
	move	d0,d1
	and	#$FF,d0
	bne	sk4		; skip this if the key has an ASCII value:
	move	d1,d0
	asr	#8,d0		; use the scancode as keycode
	bset	#7,d0		; with bit 7 set
sk4:
	cmp	#13,d0
	bne	NOTRET
RET:				; special handling for Return key
	moveq	#0,d4
	lea	9(a5),a0	; search the tree for an object that
lp5:	btst	#1,(a0)		;  is marked DEFAULT
	bne	KFOUND
	btst	#5,(a0)		; LASTOB ?
	bne	NOTRET		;  yes; check the keysub table
	addq	#1,d4		; next object
	add.w	#24,a0		
	bra	lp5

NOTRET:		; we go here if any key other than Return is pressed
	move.l	kp0,d1		; if there's no key table (null pointer)
	beq	TOP		;  then don't bother with this
	move.l	d1,a0
olp6:
	move	(a0)+,d4	; what object are we considering?
	cmp	#-1,d4
	beq	TOP		; end of list: object not found
	cmp	#-2,d4
	beq	FORCE		; end of list, but return keypress
ilp6:
	move	(a0)+,d1	; get next keycode from list
	cmp	#-1,d1
	beq	olp6		; exit if end of key list for this obj
	cmp	d1,d0
	bne	ilp6		; move on if they don't match
KFOUND:
	bsr	getstate	; let's see about this object
	btst	#3,d6		; move on if disabled
	bne	ilp6
	btst	#0,d5		; move on if not selectable
	beq	ilp6
	btst	#4,d5
	bne	RADIO		; if it's not a RADIO object,
	bsr	inv_ob		;  invert it.
	bra	RELEASED	; act as if we've clicked it
FORCE:
	bset	#14,d0		; return keycode with bit 14 set
	bra	FRCEXIT


gm:
	pea	shift
	pea	but
	pea	my
	pea	mx
	jsr	_graf_mkstate
	add.w	#16,sp
	rts

set_ob:
	bsr	getstate
	bset	#0,d6
	bra	comn_ob
clr_ob:
	bsr	getstate
	bclr	#0,d6
	bra	comn_ob
inv_ob:
	bsr	getstate
	bchg	#0,d6
;	bra	comn_ob
comn_ob:
	move	#1,-(sp)	; redraw flag
	move	d6,-(sp)	; new state
	move	22(a5),-(sp)	; clip rectangle :
	move	20(a5),-(sp)	;  use xywh of root object
	move	18(a5),-(sp)
	move	16(a5),-(sp)
	clr	-(sp)		; reserved word
	move	d4,-(sp)	; object index
	move.l	a5,-(sp)	; tree address
	jsr	_objc_change
	add.w	#20,sp
	rts
	

getstate:	; given d4, returns d5,d6,d7
		; kills d0 only

	move	d4,d0
	bra	fkludge

find_ob:
		; requires valid mouse coordinates in mx and my
		; returns d4,d5,d6,d7
		
	move	my,-(sp)
	move	mx,-(sp)
	move	#8,-(sp)	; depth of search
	clr	-(sp)		; start object = 0
	move.l	a5,-(sp)
	jsr	_objc_find
	add.w	#12,sp
	move	d0,d4
	bmi	fout
fkludge:
	mulu	#24,d0
	move	8(a5,d0.l),d5
	move	10(a5,d0.l),d6
	move	0(a5,d0.l),d7
fout:	rts
